# pylint: disable=undefined-variable
# type: ignore[name-defined]
"""
case 1:
    - Single transaction.
    - app_1 and logicsig_1 is executed
    - logicsig_1 checks the rekeyto field
    - app_1 does not do anything.

Expected:
    - The detector should not report it as vulnerable.
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - AssetCloseTo


case 2:
    - Single transaction.
    - Application and LogicSig does not check the RekeyTo field

Expected:
    - The detector should report it as vulnerable
    Vulnerable - RekeyTo
    Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 3:
    - Single transaction.
    - Only application is called
    - has_logic_sig is set to True but the logic_sig contract is not given.

Expected:
    The detector should report the operation as vulnerable. Even if the detector does not the logic sig being executed, the operation should be considered
    vulnerable because a logic sig will be vulnerable to Rekeying.
    Vulnerable - RekeyTo
    Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 4:
    - Single transaction.
    - Application is called
    - has_logic_sig is set to False.

Expected:
    The detector should report the operation as NOT VULNERABLE. Applications are not vulnerable to rekeying.
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 5:
    - Single transaction
    - Application checks the rekeyto field
    - LogicSig checks that application is called

Expected:
    NOT VULNERABLE - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo



case 6:
    - Single transaction
    - Application access the rekeyto field using `Gtxn[Txn.group_index()].rekey_to()`.
    - LogicSig checks the application is called.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 7:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application checks the rekeyto field of 0th transaction and 1st transaction using absolute indices.
        - LogicSig checks the application id
    - Second transaction (pay)(index: 1):
        - LogicSig does not check anything.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 8:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application checks the rekeyto field of 0th transaction using absolute index.
        - LogicSig checks the application id
    - Second transaction (axfer)(index: 1):
        - LogicSig checks it's rekeyto field.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 9:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application checks the rekeyto field of 1st transaction using absolute index.
        - LogicSig checks the application id
    - Second transaction (pay)(index: 1):
        - LogicSig does not check anything.

Expected:
    Vulnerable. - RekeyTo
    The first transaction is vulnerable. The second transaction should not be reported

    Vulnerable - Fee
    The first transaction is vulnerable. The second transaction should not be reported
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 10:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - LogicSig checks the application id
    - Second transaction (pay)(index: 1):
        - LogicSig checks the fields of both transactions using absolute indices.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 11:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - has_logic_sig is True.
    - Second transaction (pay)(index: 1):
        - LogicSig checks the fields of both transactions using absolute indices

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 12:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - has_logic_sig is True.
    - Second transaction (axfer)(index: 1):
        - LogicSig checks the fields of its own transaction field.

Expected:
    Vulnerable - RekeyTo
    Only the first transaction is reported as vulnerable.

    Vulnerable - Fee
    Only the first transaction is reported as vulnerable.
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 13:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
    - Second transaction (pay)(index: 1):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo
    Second transaction is reported as vulnerable

    Vulnerable - Fee
    Second transaction is reported as vulnerable

    Vulnerable - CloseRemainderTo
    Second transaction is reported as vulnerable

    Not Vulnerable - AssetCloseTo

case 14:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - LogicSig does not check anything.
    - Second transaction (axfer)(index: 1):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo
    Both First and Second transactions are reported as vulnerable.

    Vulnerable - Fee
    Both First and Second transactions are reported as vulnerable.

    Not Vulnerable - CloseRemainderTo
    Vulnerable - AssetCloseTo
    Second transaction is reported as vulnerable

case 15:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 0th transaction and 1st transaction using absolute indices
    - Second transaction (axfer)(index: 1):
        - LogicSig does not check anything.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 16:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 0th transaction using absolute index.
    - Second transaction (axfer)(index: 1):
        - LogicSig checks it's rekeyto field.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 17:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 1st transaction using absolute index.
    - Second transaction (pay)(index: 1):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo
    The first transaction is vulnerable. The second transaction should not be reported

    Vulnerable - Fee
    The first transaction is vulnerable. The second transaction should not be reported

    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 18:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application checks the rekeyto field of 0th transaction using absolute index and 1st transaction using relative index(+4).
        - LogicSig checks the application id
    - Second transaction (axfer):
        - LogicSig does not check anything.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 19:
    - Two transactions
    - First transaction (appl):
        - Application checks the rekeyto field of 1st transaction using relative index(+4).
        - LogicSig checks the application id
    - Second transaction (pay):
        - LogicSig checks rekey field of the previous transaction using (-4) relative index.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 20:
    - Two transactions
    - First transaction (appl):
        - Application checks the rekeyto field of 1st transaction using absolute index.
        - LogicSig checks the application id
    - Second transaction (pay)(index: 1):
        - LogicSig checks rekey field of the first transaction using (+5) relative index.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 21:
    - Two transactions
    - First transaction (appl):
        - Application checks the rekeyto field of 1st transaction using relative index(+2).
        - LogicSig checks the application id
    - Second transaction (pay):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo
    The first transaction is vulnerable. The second transaction should not be reported

    Vulnerable - Fee
    The first transaction is vulnerable. The second transaction should not be reported

    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 22:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - LogicSig checks the application id
    - Second transaction (axfer):
        - LogicSig checks the field of previous transaction using relative index(-1) and its own field directly.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 23:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - has_logic_sig is True.
    - Second transaction (pay):
        - LogicSig checks the field of previous transaction using relative index(-1) and its own field directly.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 24:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - has_logic_sig is True.
    - Second transaction (pay):
        - LogicSig checks the fields of its own transaction field.

Expected:
    Vulnerable - RekeyTo.
    Only the first transaction is reported as vulnerable.

    Vulnerable - Fee.
    Only the first transaction is reported as vulnerable.

    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 25:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
    - Second transaction (axfer):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo.
    Second transaction is reported as vulnerable

    Vulnerable - Fee.
    Second transaction is reported as vulnerable

    Not Vulnerable - CloseRemainderTo
    Vulnerable - AssetCloseTo
    Second transaction is reported


case 26:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - LogicSig does not check anything.
    - Second transaction (pay):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo.
    Both First and Second transactions are reported as vulnerable.

    Vulnerable - Fee.
    Both First and Second transactions are reported as vulnerable.

    Vulnerable - CloseRemainderTo
    Second transaction is reported as vulnerable

    Not Vulnerable - AssetCloseTo

case 27:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of its own transaction and 1st transaction using relative index(+1)
    - Second transaction (pay):
        - LogicSig does not check anything.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 28:
    - Two transactions
    - First transaction (appl):
        - Application checks the rekeyto field of 1st transaction using relative index(+5).
        - LogicSig checks the application id
    - Second transaction (axfer):
        - LogicSig checks rekey field of the previous transaction using (-4) relative index.

Expected:
    Vulnerable - RekeyTo.
    Second transaction uses the wrong index. First contract will be vulnerable.

    Vulnerable - Fee
    Second transaction uses the wrong index. First contract will be vulnerable.
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 29:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 1st transaction using relative index(+3).
    - Second transaction (pay):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo.
    The first transaction is vulnerable. The second transaction should not be reported

    Vulnerable - Fee.
    The first transaction is vulnerable. The second transaction should not be reported
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 30:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application checks the rekeyto field of 0th transaction using absolute index
        - LogicSig checks the application id and field of 1st transaction using relative index(+4).
    - Second transaction (pay):
        - LogicSig does not check anything.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 31:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 1st transaction using relative index(+4).
    - Second transaction (axfer):
        - LogicSig checks rekey field of the previous transaction using (-4) relative index.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo


case 32:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 1st transaction using relative index(+5)
    - Second transaction (axfer):
        - LogicSig checks rekey field of the previous transaction using (-4) relative index.

Expected:
    Vulnerable - RekeyTo.
    Second transaction uses the wrong index. First contract will be vulnerable.

    Vulnerable - Fee.
    Second transaction uses the wrong index. First contract will be vulnerable.
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 33:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 1st transaction using absolute index.
    - Second transaction (pay)(index: 1):
        - LogicSig checks rekey field of the 0th transaction using (+5) relative index.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 34:
    - Two transactions
    - First transaction (appl):
        - Application does not check anything
        - LogicSig checks the rekeyto field of 1st transaction using relative index(+2).
    - Second transaction (pay):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo.
    The first transaction is vulnerable. The second transaction should not be reported

    Vulnerable - Fee.
    The first transaction is vulnerable. The second transaction should not be reported
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 35:
    - Two transactions
    - First transaction (appl)(index: 0):
        - Application does not check anything.
        - LogicSig checks the rekeyto field of 1st transaction using relative index(+3).
    - Second transaction (axfer):
        - LogicSig does not check anything.

Expected:
    Vulnerable - RekeyTo.
    The first transaction is vulnerable. The second transaction should not be reported

    Vulnerable - Fee
    The first transaction is vulnerable. The second transaction should not be reported
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 36:
    - Single transaction (index: 4).
    - Application checks the field using absolute index.
    - LogicSig checks the application id
    - Absolute index is given in the config

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo

case 37:
    - Single transaction (index: 1)
    - Application does not check anything.
    - Logic checks the field using absolute index.
    - Absolute index is given in the config.

Expected:
    Not Vulnerable - RekeyTo
    Not Vulnerable - Fee
    Not Vulnerable - CloseRemainderTo
    Not Vulnerable - AssetCloseTo
"""
from pathlib import Path
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import


def lsig1_case_1() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.rekey_to() == Global.zero_address(),
                    Txn.fee() < Int(10000),
                    Txn.close_remainder_to() == Global.zero_address(),
                    Txn.asset_close_to() == Global.zero_address(),
                    Txn.on_completion() != OnComplete.UpdateApplication,
                    Txn.on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_2() -> Expr:
    return Return(Int(1))


def lsig1_case_5() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_6() -> Expr:
    return lsig1_case_5()


def lsig1_case_7() -> Expr:
    return lsig1_case_5()


def lsig1_case_8() -> Expr:
    return lsig1_case_5()


def lsig1_case_9() -> Expr:
    return lsig1_case_5()


def lsig1_case_10() -> Expr:
    return lsig1_case_5()


def lsig1_case_14() -> Expr:
    return Return(Int(1))


def lsig1_case_15() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    # int 0; gtxns RekeyTo
                    Gtxn[Int(0)].rekey_to() == Global.zero_address(),
                    Gtxn[1].rekey_to() == Global.zero_address(),
                    Gtxn[Int(0)].fee() < Int(100000),
                    Gtxn[1].fee() < Int(1000),
                    Gtxn[Int(0)].close_remainder_to() == Global.zero_address(),
                    Gtxn[1].close_remainder_to() == Global.zero_address(),
                    Gtxn[Int(0)].asset_close_to() == Global.zero_address(),
                    Gtxn[1].asset_close_to() == Global.zero_address(),
                    Gtxn[0].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[0].on_completion() != OnComplete.DeleteApplication,
                    Gtxn[Int(1)].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[Int(1)].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_16() -> Expr:
    return Seq(
        [
            Assert(Gtxn[0].rekey_to() == Global.zero_address()),
            Assert(Gtxn[0].fee() < Int(10000)),
            Assert(Gtxn[0].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[0].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[0].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[0].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_17() -> Expr:
    return Seq(
        [
            Assert(Gtxn[1].rekey_to() == Global.zero_address()),
            Assert(Gtxn[1].fee() < Int(10000)),
            Assert(Gtxn[1].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[1].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[1].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[1].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_18() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_19() -> Expr:
    return lsig1_case_18()


def lsig1_case_20() -> Expr:
    return lsig1_case_18()


def lsig1_case_21() -> Expr:
    return lsig1_case_18()


def lsig1_case_22() -> Expr:
    return lsig1_case_18()


def lsig1_case_26() -> Expr:
    return Return(Int(1))


def lsig1_case_27() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(1)].rekey_to() == Global.zero_address(),
                    Txn.fee() < Int(10000),
                    Gtxn[Txn.group_index() + Int(1)].fee() < Int(10000),
                    Txn.close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(1)].close_remainder_to() == Global.zero_address(),
                    Txn.asset_close_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(1)].asset_close_to() == Global.zero_address(),
                    Txn.on_completion() != OnComplete.UpdateApplication,
                    Txn.on_completion() != OnComplete.DeleteApplication,
                    Gtxn[Txn.group_index() + Int(1)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(1)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_28() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_29() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(3)].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(3)].fee() < Int(10000),
                    Gtxn[Txn.group_index() + Int(3)].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(3)].asset_close_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(3)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(3)].on_completion()
                    != OnComplete.UpdateApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_30() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                    Gtxn[Txn.group_index() + Int(4)].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(4)].fee() < Int(10000),
                    Gtxn[Txn.group_index() + Int(4)].asset_close_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(4)].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(4)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(4)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_31() -> Expr:
    return lsig1_case_30()


def lsig1_case_32() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                    Gtxn[Txn.group_index() + Int(5)].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(5)].fee() < Int(10000),
                    Gtxn[Txn.group_index() + Int(5)].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(5)].asset_close_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_33() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[1].rekey_to() == Global.zero_address(),
                    Gtxn[1].fee() < Int(10000),
                    Gtxn[1].close_remainder_to() == Global.zero_address(),
                    Gtxn[1].asset_close_to() == Global.zero_address(),
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                    Gtxn[1].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[1].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_34() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(2)].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(2)].fee() < Int(10000),
                    Gtxn[Txn.group_index() + Int(2)].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(2)].asset_close_to() == Global.zero_address(),
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                    Gtxn[Txn.group_index() + Int(2)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(2)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_35() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(3)].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(3)].fee() < Int(10000),
                    Gtxn[Txn.group_index() + Int(3)].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(3)].asset_close_to() == Global.zero_address(),
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                    Gtxn[Txn.group_index() + Int(3)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(3)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_36() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.application_id() == Int(1337),
                    # Txn.on_completion() == OnComplete.NoOp,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig1_case_37() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[1].fee() < Int(10000),
                    Gtxn[1].rekey_to() == Global.zero_address(),
                    Gtxn[1].close_remainder_to() == Global.zero_address(),
                    Gtxn[1].asset_close_to() == Global.zero_address(),
                    Gtxn[1].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[1].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def logic_sig() -> Expr:
    test_case = Btoi(Arg(0))
    return Seq(
        [
            Cond(
                [test_case == Int(1), lsig1_case_1()],
                [test_case == Int(2), lsig1_case_2()],
                # [test_case == Int(3), lsig1_case_3()],
                # [test_case == Int(4), lsig1_case_4()],
                [test_case == Int(5), lsig1_case_5()],
                [test_case == Int(6), lsig1_case_6()],
                [test_case == Int(7), lsig1_case_7()],
                [test_case == Int(8), lsig1_case_8()],
                [test_case == Int(9), lsig1_case_9()],
                [test_case == Int(10), lsig1_case_10()],
                # [test_case == Int(11), lsig1_case_11()],
                # [test_case == Int(12), lsig1_case_12()],
                # [test_case == Int(13), lsig1_case_13()],
                [test_case == Int(14), lsig1_case_14()],
                [test_case == Int(15), lsig1_case_15()],
                [test_case == Int(16), lsig1_case_16()],
                [test_case == Int(17), lsig1_case_17()],
                [test_case == Int(18), lsig1_case_18()],
                [test_case == Int(19), lsig1_case_19()],
                [test_case == Int(20), lsig1_case_20()],
                [test_case == Int(21), lsig1_case_21()],
                [test_case == Int(22), lsig1_case_22()],
                # [test_case == Int(23), lsig1_case_23()],
                # [test_case == Int(24), lsig1_case_24()],
                # [test_case == Int(25), lsig1_case_25()],
                [test_case == Int(26), lsig1_case_26()],
                [test_case == Int(27), lsig1_case_27()],
                [test_case == Int(28), lsig1_case_28()],
                [test_case == Int(29), lsig1_case_29()],
                [test_case == Int(30), lsig1_case_30()],
                [test_case == Int(31), lsig1_case_31()],
                [test_case == Int(32), lsig1_case_32()],
                [test_case == Int(33), lsig1_case_33()],
                [test_case == Int(34), lsig1_case_34()],
                [test_case == Int(35), lsig1_case_35()],
                [test_case == Int(36), lsig1_case_36()],
                [test_case == Int(37), lsig1_case_37()],
            )
        ]
    )


def compile_program(output_file: Path) -> None:
    teal = compileTeal(logic_sig(), mode=Mode.Signature, version=7)
    with open(output_file, "w", encoding="utf-8") as f:
        f.write(teal)


if __name__ == "__main__":
    compile_program(Path("logicsig_1.teal"))



"""
NOTE: Incorrect or Inaccurate documentation. ignore for now.

tealer init

positional_arguments: (+), list of filesystem paths

optional_argument: --group-config yaml file

tealer detect



subcommands

- `init`
- `detect`
- `print`

#### init

positional arguments:
    programs        list of space-separated directories/files.

optional argument:
    --group-config  yaml configuration file.

Purpose:

program_files = []
for program in programs:
    if program is directory:
        list all files with .teal extension and add them to program_files.
    if program is file:
        add it to program_files

if group_config is not null:
    read the config
else:
    create a empty config

if a program file is already in group_config:
    remove it from program_files

For each program in program_files:
    analyze the program and append its information to the group_config.

Output updated group-config file
"""

import sys

# from pathlib import Path
import inspect
import argparse

from typing import List, Type, Tuple, TYPE_CHECKING, Dict
from pkg_resources import iter_entry_points  # type: ignore

from tealer.detectors.abstract_detector import (
    AbstractDetector,
)
from tealer.printers.abstract_printer import AbstractPrinter
from tealer.detectors import all_detectors
from tealer.printers import all_printers
from tealer.exceptions import TealerException

# from tealer.utils.command_line.group_config import (
#     read_config_from_file,
# )
from tealer.teal.parse_teal import parse_teal
from tealer.teal.parse_functions import construct_function
from tealer.utils.teal_enums import ContractType, contract_type_from_txt
from tealer.tealer import Tealer

# from tealer.utils.output import subroutine_to_dot
from tealer.execution_context.transactions import (
    Transaction,
    GroupTransaction,
    fill_group_relative_indexes,
)
from tealer.utils.command_line.group_config import USER_CONFIG_TRANSACTION_TYPES


if TYPE_CHECKING:
    from tealer.teal.functions import Function
    from tealer.teal.teal import Teal
    from tealer.utils.command_line.group_config import (
        GroupConfig,
        GroupConfigFunctionCall,
    )


def collect_plugins() -> Tuple[List[Type[AbstractDetector]], List[Type[AbstractPrinter]]]:
    """collect detectors and printers installed in form of plugins.

    plugins are collected using the entry point group `teal_analyzer.plugin`.
    The entry point of each plugin has to return tuple containing list of detectors and
    list of printers defined in the plugin when called.

    Returns:
        (Tuple[List[Type[AbstractDetector]], List[Type[AbstractPrinter]]]): detectors and
        printers added in the form of plugins.

    Raises:
        TealerException: raises exception if a plugin's detector is not a subclass of AbstractDetector
            or if a plugin's printer is not a subclass of AbstractPrinter.
    """
    detector_classes: List[Type[AbstractDetector]] = []
    printer_classes: List[Type[AbstractPrinter]] = []
    for entry_point in iter_entry_points(group="teal_analyzer.plugin", name=None):
        make_plugin = entry_point.load()

        plugin_detectors, plugin_printers = make_plugin()
        detector = None
        if not all(issubclass(detector, AbstractDetector) for detector in plugin_detectors):
            raise TealerException(
                f"Error when loading plugin {entry_point}, {detector} is not a detector"
            )
        printer = None
        if not all(issubclass(printer, AbstractPrinter) for printer in plugin_printers):
            raise TealerException(
                f"Error when loading plugin {entry_point}, {printer} is not a printer"
            )

        detector_classes += plugin_detectors
        printer_classes += plugin_printers

    return detector_classes, printer_classes


def get_detectors_and_printers() -> Tuple[
    List[Type[AbstractDetector]], List[Type[AbstractPrinter]]
]:
    """Get list of detectors and printers available to tealer.

    Detectors, Printers are considered available to tealer either if they
    are defined in the tealer itself or if they are defined in one of the
    tealer plugins installed in the system.

    Returns:
        list of detectors, list of printers defined in tealer and plugins
        combined.
    """

    detector_classes = [getattr(all_detectors, name) for name in dir(all_detectors)]
    detector_classes = [
        d for d in detector_classes if inspect.isclass(d) and issubclass(d, AbstractDetector)
    ]

    printer_classes = [getattr(all_printers, name) for name in dir(all_printers)]
    printer_classes = [
        d for d in printer_classes if inspect.isclass(d) and issubclass(d, AbstractPrinter)
    ]

    plugins_detectors, plugins_printers = collect_plugins()

    detector_classes += plugins_detectors
    printer_classes += plugins_printers

    return detector_classes, printer_classes


# pylint: disable=too-many-branches
def validate_command_line_options(args: argparse.Namespace) -> None:
    """Validate command line options and print message and exit the program

    - Only one of --init or --detect or --print must be selected
    - init
        - --contracts is required
        - --group-config is optional
    - detect
        - Only one of --contracts or --group_config is selected

    Args:
        args: Command line arguments
    """

    def print_and_exit(message: str) -> None:
        print(f"CommandLineError: {message}")
        sys.exit(1)

    if not args.subcommand:
        print_and_exit("Use one of these subcommand: detect | print  | regex")

    if args.subcommand == "detect":
        if args.contracts is None and args.group_config is None:
            print_and_exit("--detect requires one of --contracts or --group-config")
        if args.contracts is not None and args.group_config is not None:
            print_and_exit(
                "--detect takes only one of --contracts or --group-config. Use --group-config with --init to add new contracts to existing configuration."
            )
    elif args.subcommand == "print" and args.printers_to_run is not None:
        if args.contracts is None and args.group_config is None:
            print_and_exit("--print requires one of --contracts or --group-config")
        if args.contracts is not None and args.group_config is not None:
            print_and_exit(
                "--print takes only one of --contracts or --group-config. Use --group-config with --init to add new contracts to existing configuration."
            )

        if args.json is not None:
            print_and_exit("--json is not supported for --print.")

    # TODO: update args.network. args.network defaults to mainnet and is never None.
    # if args.network is not None and args.contracts is None:
    #     print_and_exit("--network is not supported when --contracts is not supplied.")

    if args.subcommand == "detect" and args.detectors_to_run is None:
        if (
            args.detectors_to_exclude is not None
            or args.exclude_stateless
            or args.exclude_stateful
            or args.filter_paths is not None
        ):
            print_and_exit(
                "--exclude, --exclude-stateless, --exclude-stateful and --filter-paths options are only available when --detect is selected."
            )


# def handle_detect(args: argparse.Namespace) -> None:
# group_config = read_config_from_file(Path(args.group_config))
# tealer = init_tealer_from_config(group_config)

# for contract_name, contract in tealer.contracts.items():
#     for function_name, function in contract.functions.items():
#         print("None")
# dot_file_name = f"{contract_name}_{function_name}.dot"
# with open(dot_file_name, "w", encoding="utf-8") as f:
#     f.write(subroutine_to_dot(function.main))

# output = tealer


# def list_contract_files(directory: Path) -> List[Path]:
#     teal_files = []
#     for file in directory.iterdir():
#         if file.is_file() and file.name.endswith(".teal"):
#             teal_files.append(directory / file)
#     return teal_files


def _get_function_from_config(
    function_call_config: "GroupConfigFunctionCall", contracts: Dict[str, "Teal"]
) -> "Function":
    if function_call_config.contract not in contracts:
        raise TealerException(f"{function_call_config.contract} not found in listed contracts")
    contract_obj = contracts[function_call_config.contract]
    if function_call_config.function not in contract_obj.functions:
        raise TealerException(
            f"{function_call_config.function} not found in {contract_obj.contract_name} functions."
        )
    return contract_obj.functions[function_call_config.function]


# pylint: disable=too-many-locals
def init_tealer_from_config(config: "GroupConfig") -> "Tealer":
    contracts: Dict[str, "Teal"] = {}
    for contract_config in config.contracts:
        with open(contract_config.file_path, encoding="utf-8") as f:
            teal = parse_teal(f.read(), contract_config.name)
        given_contract_type = contract_type_from_txt(contract_config.contract_type)
        teal.contract_type = given_contract_type
        # ignore version and subroutines
        contract_functions: Dict[str, "Function"] = {}
        for function_config in contract_config.functions:
            func = construct_function(teal, function_config.dispatch_path, function_config.name)
            contract_functions[function_config.name] = func

        teal.functions = contract_functions
        contracts[contract_config.name] = teal

    group_objs_list: List[GroupTransaction] = []
    for txn_config in config.groups:
        txn_id_to_obj: Dict[str, "Transaction"] = {}
        for txn in txn_config.transactions:
            txn_obj = Transaction()
            txn_obj.transacton_id = txn.txn_id
            txn_obj.type = USER_CONFIG_TRANSACTION_TYPES[txn.txn_type]
            if txn.has_logic_sig is not None:
                txn_obj.has_logic_sig = txn.has_logic_sig
            txn_obj.absoulte_index = txn.absolute_index

            if txn.application is not None:
                app_function = _get_function_from_config(txn.application, contracts)
                if app_function.contract.contract_type == ContractType.LogicSig:
                    raise TealerException(
                        f"{txn.application.contract} is a logic-sig but is given as an application."
                    )
                txn_obj.application = app_function

            if txn.logic_sig is not None:
                logic_sig_function = _get_function_from_config(txn.logic_sig, contracts)
                if logic_sig_function.contract.contract_type != ContractType.LogicSig:
                    raise TealerException(
                        f"{txn.logic_sig.contract} is an application but is given as a logic-sig."
                    )
                txn_obj.logic_sig = logic_sig_function
                txn_obj.has_logic_sig = True

            if txn.txn_id in txn_id_to_obj:
                raise TealerException(f"{txn.txn_id} is repeated in the same group.")
            txn_id_to_obj[txn.txn_id] = txn_obj

        group_obj = GroupTransaction()
        group_obj.operation_name = txn_config.operation
        group_obj.transactions = list(txn_id_to_obj.values())

        for txn in txn_config.transactions:
            txn_obj = txn_id_to_obj[txn.txn_id]
            txn_obj.group_transaction = group_obj
            if txn.relative_indexes is not None:
                for other_txn_id in txn.relative_indexes:
                    if other_txn_id not in txn_id_to_obj:
                        raise TealerException(
                            f"other_txn_id: {other_txn_id} is not present in the same group"
                        )
                    offset = txn.relative_indexes[other_txn_id]
                    txn_obj.relative_indexes[offset] = txn_id_to_obj[other_txn_id]

            if txn_obj.absoulte_index is not None:
                if txn_obj.absoulte_index in group_obj.absolute_indexes:
                    raise TealerException(
                        f"Two transactions have same absolute index {txn_obj.transacton_id}, {group_obj.absolute_indexes[txn_obj.absoulte_index].transacton_id}"
                    )
                group_obj.absolute_indexes[txn_obj.absoulte_index] = txn_obj

        fill_group_relative_indexes(group_obj)
        group_objs_list.append(group_obj)

    return Tealer(contracts, group_objs_list, output_group=True)


def init_tealer_from_single_contract(contract_src: str, contract_name: str) -> "Tealer":
    teal = parse_teal(contract_src, contract_name)
    contracts: Dict[str, "Teal"] = {contract_name: teal}
    contract_functions: Dict[str, "Function"] = {}
    contract_functions[contract_name] = construct_function(teal, ["B0"], contract_name)
    teal.functions = contract_functions

    txn_obj = Transaction()
    if teal.contract_type == ContractType.LogicSig:
        txn_obj.transacton_id = contract_name
        txn_obj.has_logic_sig = True
        txn_obj.logic_sig = contract_functions[contract_name]
    else:
        # contract is either Approval or ClearState
        txn_obj.application = contract_functions[contract_name]

    group_obj = GroupTransaction()
    group_obj.operation_name = contract_name
    group_obj.transactions = [txn_obj]
    group_objs_list = [group_obj]

    return Tealer(contracts, group_objs_list)



"""Defines classes to represent asset_holding_get fields.

``asset_holding_get`` instruction is used to access holding fields
of an asset for the given account.

Each field that can be accessed using asset_holding_get is represented
by a class in tealer. All the classes representing the fields must inherit
from AssetHoldingField class.

"""

# pylint: disable=too-few-public-methods
class AssetHoldingField:
    """Base class to represent asset_holding_get field."""

    def __init__(self) -> None:
        self._version: int = 2

    @property
    def version(self) -> int:
        """Teal version this field is introduced in and supported from.

        Returns:
            Teal version the field is introduced in and supported from.
        """
        return self._version

    def __str__(self) -> str:
        return self.__class__.__qualname__


class AssetBalance(AssetHoldingField):
    """Amount of the asset unit held by this account."""


class AssetFrozen(AssetHoldingField):
    """Is asset frozen for this account or not."""



"""Parser for fields of app_params_get.

Each ``app_params_get`` field is represented as a class. Parsing
the field is creating the class instance representing the field given
it's string representation.

Fields of ``app_params_get`` doesn't have immediate arguments
and have very simple structure. All fields map one-to-one with
their string representation. As a result, app parameter fields are
parsed by first constructing a map(dict) from their string
representation to corresponding class representing the field and
using that to find the correct class for the given field.

Attributes:
    APP_PARAMS_FIELD_TXT_TO_OBJECT: Map(dict) from string representation
        of app parameter field to the corresponding class.
"""

from tealer.teal.instructions.app_params_field import (
    AppParamsField,
    AppApprovalProgram,
    AppClearStateProgram,
    AppGlobalNumUint,
    AppGlobalNumByteSlice,
    AppLocalNumUint,
    AppLocalNumByteSlice,
    AppExtraProgramPages,
    AppCreator,
    AppAddress,
)


APP_PARAMS_FIELD_TXT_TO_OBJECT = {
    "AppParamsField": AppParamsField,
    "AppApprovalProgram": AppApprovalProgram,
    "AppClearStateProgram": AppClearStateProgram,
    "AppGlobalNumUint": AppGlobalNumUint,
    "AppGlobalNumByteSlice": AppGlobalNumByteSlice,
    "AppLocalNumUint": AppLocalNumUint,
    "AppLocalNumByteSlice": AppLocalNumByteSlice,
    "AppExtraProgramPages": AppExtraProgramPages,
    "AppCreator": AppCreator,
    "AppAddress": AppAddress,
}


def parse_app_params_field(field: str) -> AppParamsField:
    """Parse fields of ``app_params_get``.

    Args:
        field: string representation of the field.

    Returns:
        object of class corresponding to the given app_params_get field.
    """

    return APP_PARAMS_FIELD_TXT_TO_OBJECT[field]()



"""Detector for finding unoptimized access to the sender."""

from typing import TYPE_CHECKING, List

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.teal.instructions.instructions import Instruction, Txna
from tealer.teal.instructions.transaction_field import Accounts
from tealer.utils.output import InstructionsOutput

if TYPE_CHECKING:
    from tealer.utils.output import ListOutput


class SenderAccess(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find unoptimized access to the sender"""

    NAME = "sender-access"
    DESCRIPTION = "Unoptimized Gtxn"
    TYPE = DetectorType.STATELESS

    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = (
        "https://github.com/crytic/tealer/wiki/Detector-Documentation#Unoptimized-sender-access"
    )
    WIKI_TITLE = "Unoptimized sender access"
    WIKI_DESCRIPTION = "Txn.accounts[0] can be optimized to be Txn.sender()]"
    WIKI_EXPLOIT_SCENARIO = """"""

    WIKI_RECOMMENDATION = """Use Txn.sender()."""

    def detect(self) -> "ListOutput":
        """Detector to find unoptimized access to the sender

        Returns:
              The instruction to optimized
        """

        detector_output: "ListOutput" = []

        for contract in self.tealer.contracts.values():

            all_findings: List[List[Instruction]] = []

            for function in contract.functions.values():
                for block in function.blocks:
                    for ins in block.instructions:
                        if isinstance(ins, Txna) and isinstance(ins.field, Accounts) and ins.field.idx == 0:  # type: ignore
                            all_findings.append([ins])

            if all_findings:
                detector_output.append(InstructionsOutput(contract, self, all_findings))

        return detector_output



"""Defines abstract base class for printers in tealer.

This module contains implementation of abstract class which
defines the attributes, common methods and must implement methods
of printers. All printers in tealer must inherit from this class
and set attributes, override abstract methods.

Classes:
    AbstractPrinter: Abstract class to represent printers in tealer.
"""

import abc

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tealer.teal.teal import Teal
    from pathlib import Path


class IncorrectPrinterInitialization(Exception):
    """Exception class to represent incorrect printer intialization.

    This exception will be used if any of the necessary attributes
    of the AbstractPrinter are not set by the inheriting printer class.
    """


class AbstractPrinter(metaclass=abc.ABCMeta):  # pylint: disable=too-few-public-methods
    """Abstract class to represent printers in tealer.

    All printers in tealer must inherit from this class and override the
    abstract methods with functionality specific to them. This class defines
    the api methods that must be implemented by every printers. ``print``
    method of each printer will be called to execute the printer. Every
    printer must implement ``print`` method.

    Attrributes:
        NAME: Name of the printer. printer name will be used while
            displaying information about the printer and also for selecting
            the printer from command line.
        HELP: Description of the printer.

    Args:
        teal: Teal instance representing the contract. Printer will run on
            the contract represented by :teal:

    Raises:
        IncorrectPrinterInitialization: Exception is raised if NAME, HELP
            attributes are not set by the inheriting printer class.
    """

    NAME = ""
    HELP = ""
    WIKI_URL = ""

    def __init__(self, teal: "Teal"):
        self.teal = teal

        if not self.NAME:
            raise IncorrectPrinterInitialization(
                f"NAME is not initialized {self.__class__.__name__}"
            )

        if not self.HELP:
            raise IncorrectPrinterInitialization(
                f"HELP is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_URL:
            raise IncorrectPrinterInitialization(
                f"WIKI_URL is not initialized for {self.__class__.__name__}"
            )

    @abc.abstractmethod
    def print(self) -> None:
        """entry method of the printer.

        All printers must override this method with the functionality
        specific to them. This method is the entry point of the printer
        and will be called to execute it.
        """



# pylint: skip-file
# mypy: ignore-errors
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import

from pyteal import *

from tealer.detectors.all_detectors import IsUpdatable, IsDeletable

router = Router(
    name="Example",
    bare_calls=BareCallActions(),
)

"""
Pattern for method config having CallConfig.CALL:
    txn OnCompletion
    int NoOp
    ==
    txn ApplicationID
    int 0
    !=
    &&
    assert

Pattern for method config having CallConfig.CREATE:
    txn OnCompletion
    int NoOp
    ==
    txn ApplicationID
    int 0
    ==
    &&
    assert

Pattern for method config having CallConfig.ALL:
    txn OnCompletion
    int NoOp
    ==
    assert

Tealer cannot calculate the information from first two patterns. For that reason, all of the
methods have CallConfig.ALL.
"""


@router.method(no_op=CallConfig.ALL)
def echo(input: abi.Uint64, *, output: abi.Uint64) -> Expr:
    """
    Method config validations Teal pattern:
        txn OnCompletion
        int NoOp
        ==
        txn ApplicationID
        int 0
        !=
        &&
        assert            // Assert(NoOp, CALL)

    Args:
        input: Input argument
        output: Return value of the echo method.

    Returns:
        Body of the echo method as PyTeal expr.
    """
    return output.set(input.get())


@router.method(opt_in=CallConfig.ALL)
def deposit() -> Expr:
    return Return()


@router.method(close_out=CallConfig.ALL)
def getBalance() -> Expr:
    return Return()


@router.method(update_application=CallConfig.ALL)
def update() -> Expr:
    return Err()


@router.method(delete_application=CallConfig.ALL)
def delete() -> Expr:
    return Err()


# does not matter even if multiple actions are allowed as it fails for all calls.
@router.method(no_op=CallConfig.CALL, opt_in=CallConfig.CREATE, close_out=CallConfig.ALL)
def some(input: abi.Uint64) -> Expr:
    return Err()


# PyTeal creates intcblock, bytecblock if assemble_constants = True
# int NoOp, OptIn, ... are all replaced by intc_* instructions.
# pragma(compiler_version="0.22.0")
approval_program, clear_state_program, contract = router.compile_program(
    version=7,
    assemble_constants=True,  # use intcblock, bytecblock
    # optimize=OptimizeOptions(scratch_slots=True),
)

router_with_assembled_constants = [
    (approval_program, IsUpdatable, []),
    (approval_program, IsDeletable, []),
]



"""Parser for global fields.

Each ``global`` field is represented as a class. Parsing
the field is creating the class instance representing the field given
it's string representation.

Fields of ``global`` doesn't have immediate arguments
and have very simple structure. All fields map one-to-one with
their string representation. As a result, global fields are
parsed by first constructing a map(dict) from their string
representation to corresponding class representing the field and
using that to find the correct class for the given field.

Attributes:
    GLOBAL_FIELD_TXT_TO_OBJECT: Map(dict) from string representation
        of global field to the corresponding class.
"""

from tealer.teal.global_field import (
    GroupSize,
    ZeroAddress,
    MinTxnFee,
    GlobalField,
    MinBalance,
    MaxTxnLife,
    LogicSigVersion,
    Round,
    LatestTimestamp,
    CurrentApplicationID,
    CreatorAddress,
    CurrentApplicationAddress,
    GroupID,
    OpcodeBudget,
    CallerApplicationID,
    CallerApplicationAddress,
)

GLOBAL_FIELD_TXT_TO_OBJECT = {
    "GroupSize": GroupSize,
    "MinTxnFee": MinTxnFee,
    "ZeroAddress": ZeroAddress,
    "MinBalance": MinBalance,
    "MaxTxnLife": MaxTxnLife,
    "LogicSigVersion": LogicSigVersion,
    "Round": Round,
    "LatestTimestamp": LatestTimestamp,
    "CurrentApplicationID": CurrentApplicationID,
    "CreatorAddress": CreatorAddress,
    "CurrentApplicationAddress": CurrentApplicationAddress,
    "GroupID": GroupID,
    "OpcodeBudget": OpcodeBudget,
    "CallerApplicationID": CallerApplicationID,
    "CallerApplicationAddress": CallerApplicationAddress,
}


def parse_global_field(field: str) -> GlobalField:
    """Parse global fields.

    Args:
        field: string representation of the field.

    Returns:
        object of class corresponding to the given global field.
    """

    return GLOBAL_FIELD_TXT_TO_OBJECT[field]()



from typing import List, Tuple, Union
import pytest

from tealer.utils.command_line.common import init_tealer_from_single_contract
from tealer.utils.algorand_constants import MAX_UINT64, MAX_GROUP_SIZE

from tests.utils import order_basic_blocks

BASIC = """
#pragma version 6
txn Fee
int 1000
<=
assert
txn GroupIndex
int 0
==
bz fail_txn
callsub perform_validations
process_txn:
int 1
return
fail_txn:
int 0
return
perform_validations:
int 1
assert
retsub
"""

txn_fees = [1000, 1000, 1000, 0, 1000]
gtxn_fees = [[1000, 1000, 1000, 0, 1000], *[[0] * 5 for _ in range(MAX_GROUP_SIZE - 1)]]

BASIC_LISTS = [txn_fees, gtxn_fees]

BASIC_2 = """
#pragma version 6
txn Fee
global MinTxnFee
<=
assert
txn GroupIndex
int 0
==
bz fail_txn
callsub perform_validations
process_txn:
int 1
return
fail_txn:
int 0
return
perform_validations:
int 1
assert
retsub
"""

txn_fees = [True, True, True, 0, True]
gtxn_fees = [[True, True, True, 0, True], *[[0] * 5 for _ in range(MAX_GROUP_SIZE - 1)]]

BASIC_2_LISTS = [txn_fees, gtxn_fees]

NO_CHECKS = """
#pragma version 6
callsub perform_validations
process_txn:
int 1
return
perform_validations:
int 1
assert
retsub
"""

txn_fees = [MAX_UINT64] * 3
gtxn_fees = [[MAX_UINT64] * 3 for _ in range(MAX_GROUP_SIZE)]

NO_CHECKS_LISTS = [txn_fees, gtxn_fees]

GTXN_0 = """
#pragma version 6
txn GroupIndex
int 0
==
bz fail_txn

gtxn 0 Fee
int 1000
>=
bnz fail_txn

callsub perform_validations

process_txn:
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub
"""

txn_fees = [MAX_UINT64] * 6
txn_fees[4] = 0  # fail_txn block
gtxn_fees = [
    [
        999,
        999,
        999,
        999,
        0,
        999,
    ],
    *[[0] * 6 for _ in range(MAX_GROUP_SIZE - 1)],
]

GTXN_0_LISTS = [txn_fees, gtxn_fees]

CHECK_IN_SUBROUTINE = """
#pragma version 6
callsub validate_fee
callsub perform_validations

process_txn:
int 1
return

fail_txn:
int 0
return

perform_validations:
int 1
assert
retsub

validate_fee:
    txn Fee
    int 1000
    >
    bnz fail_txn
    retsub
"""

txn_fees = [1000, 1000, 1000, 0, 1000, 1000, 1000]
gtxn_fees = [list(txn_fees) for _ in range(MAX_GROUP_SIZE)]

CHECK_IN_SUBROUTINE_LISTS = [txn_fees, gtxn_fees]

MULTIPLE_GROUP_SIZES_IN_SUBROUTINES = """
#pragma version 6
txn Fee
int 10000
>=
bnz fail_txn
callsub validate_fee
callsub perform_validations

process_txn:
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub

validate_fee:
    global GroupSize
    int 2
    ==
    bnz group_size_2

    global GroupSize
    int 1
    ==
    assert
    gtxn 0 Fee
    int 10000
    <=
    assert
    retsub

    group_size_2:
        gtxn 0 Fee
        int 1000
        <
        assert
        gtxn 1 Fee
        int 1000
        <
        bz fail_txn

    retsub
"""

"""
#pragma version 6
txn Fee
int 10000
>=
bnz fail_txn
callsub validate_fee
callsub perform_validations
int 1 return

fail_txn:
    int 0
    return

perform_validations:
    b fail_txn

validate_fee:
    global GroupSize
    int 2
    ==
    bnz group_size_2

    global GroupSize
    int 1
    ==
    assert
    gtxn 0 Fee
    int 10000
    <=
    assert
    retsub

    group_size_2:
        gtxn 0 Fee
        int 1000
        <
        assert
        gtxn 1 Fee
        int 1000
        <
        bz fail_txn

    retsub
"""

txn_fees = [9999] * 10
txn_fees[4] = 0  # fail_txn block
gtxn_fees = [
    [9999, 9999, 9999, 9999, 0, 9999, 9999, 9999, 999, 999],
    [999, 999, 999, 999, 0, 999, 999, 0, 999, 999, 999],
    *[[0] * 10 for _ in range(MAX_GROUP_SIZE - 2)],
]


MULTIPLE_GROUP_SIZES_IN_SUBROUTINES_LISTS = [txn_fees, gtxn_fees]


CHECKS_IN_MULTIPLE_SUBROUTINES = """
#pragma version 6
global GroupSize
int 2
==
bnz group_size_2

global GroupSize
int 1
==
assert
callsub validate_fee_group_size_1
b process_txn

group_size_2:
    callsub validate_fee_group_size_2
    b process_txn

process_txn:
    callsub perform_validations
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub

validate_fee_group_size_1:
    gtxn 0 Fee
    int 10000
    <
    assert
    retsub

validate_fee_group_size_2:
    gtxn 0 Fee
    int 1000
    <=
    assert
    gtxn 1 Fee
    int 1000
    <=
    bz fail_txn
    retsub
"""

txn_fees = [MAX_UINT64] * 12
txn_fees[7] = 0  # fail_txn block
gtxn_fees = [
    # [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    [9999, 9999, 9999, 1000, 1000, 9999, 9999, 0, 9999, 9999, 1000, 1000],
    [1000, 0, 0, 1000, 1000, 1000, 1000, 0, 1000, 0, 1000, 1000, 1000],
    *[[0] * 12 for _ in range(MAX_GROUP_SIZE - 2)],
]


CHECKS_IN_MULTIPLE_SUBROUTINES_LISTS = [txn_fees, gtxn_fees]

CHECKS_IN_MULTIPLE_SUBROUTINES_2 = """
#pragma version 6
global GroupSize
int 2
==
bnz group_size_2

global GroupSize
int 1
==
assert
callsub validate_fee_group_size_1
b process_txn

group_size_2:
    callsub validate_fee_group_size_2
    b process_txn

process_txn:
    callsub perform_validations
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub

validate_fee_group_size_1:
    gtxn 0 Fee
    int 10000
    <
    assert
    retsub

validate_fee_group_size_2:
    gtxn 0 Fee
    global MinTxnFee
    int 3
    *
    <=                  // detection of this pattern is possible because of StackValue emulation/analysis
    assert
    gtxn 1 Fee
    int 1000
    int 4
    *
    <=
    bz fail_txn
    retsub
"""

txn_fees = [MAX_UINT64] * 12
txn_fees[7] = 0  # fail_txn block
gtxn_fees = [
    # [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    [True, 9999, 9999, True, True, True, True, 0, True, 9999, True, True],
    [True, 0, 0, True, True, True, True, 0, True, 0, True, True, True],
    *[[0] * 12 for _ in range(MAX_GROUP_SIZE - 2)],
]

CHECKS_IN_MULTIPLE_SUBROUTINES_LISTS_2 = [txn_fees, gtxn_fees]


CHECK_IN_ONE_PATH = """
#pragma version 6
txn RekeyTo
global ZeroAddress
==
bnz rekey_zero

global GroupSize
int 1
==
assert
txn Fee
int 0
==
assert
b process_txn

rekey_zero:
b process_txn

process_txn:
    callsub perform_validations
    int 1
    return

perform_validations:
    int 1
    assert
    retsub
"""

txn_fees = [MAX_UINT64] * 6
txn_fees[1] = 0
gtxn_fees = [list(txn_fees) for _ in range(MAX_GROUP_SIZE)]


CHECK_IN_ONE_PATH_LISTS = [txn_fees, gtxn_fees]


ALL_TESTS = [
    (BASIC, *BASIC_LISTS),
    (NO_CHECKS, *NO_CHECKS_LISTS),
    (GTXN_0, *GTXN_0_LISTS),
    (CHECK_IN_SUBROUTINE, *CHECK_IN_SUBROUTINE_LISTS),
    # (MULTIPLE_GROUP_SIZES_IN_SUBROUTINES, *MULTIPLE_GROUP_SIZES_IN_SUBROUTINES_LISTS),
    (CHECKS_IN_MULTIPLE_SUBROUTINES, *CHECKS_IN_MULTIPLE_SUBROUTINES_LISTS),
    (CHECK_IN_ONE_PATH, *CHECK_IN_ONE_PATH_LISTS),
    (BASIC_2, *BASIC_2_LISTS),
    (CHECKS_IN_MULTIPLE_SUBROUTINES_2, *CHECKS_IN_MULTIPLE_SUBROUTINES_LISTS_2),
]


@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_tx_types_gtxn(
    test: Tuple[str, List[Union[bool, int]], List[List[Union[bool, int]]]]
) -> None:
    code, max_fees_list, gtxn_fees_list = test

    tealer = init_tealer_from_single_contract(code.strip(), "test")
    function = tealer.contracts["test"].functions["test"]

    bbs = order_basic_blocks(function.blocks)
    for i, b in enumerate(bbs):
        print("X:", i, repr(b), function.transaction_context(b).max_fee)
        if isinstance(max_fees_list[i], bool):
            assert function.transaction_context(b).max_fee_unknown is True
        else:
            assert function.transaction_context(b).max_fee == max_fees_list[i]
        for ind in range(MAX_GROUP_SIZE):
            ctx = function.transaction_context(b).gtxn_context(ind)
            print(f"Y: {ind}, {i}", ctx.max_fee, gtxn_fees_list[ind][i])
            if isinstance(gtxn_fees_list[ind][i], bool):
                assert ctx.max_fee_unknown is True
            else:
                assert ctx.max_fee == gtxn_fees_list[ind][i]



"""Defines a class to represent a Function in Teal contract.

Functions are different from subroutine. Functions are abstract blocks of code, each callable
by a end-user. Function represents an operation of the contract.

Classes:
    Function: Represents a Function.
"""

from typing import TYPE_CHECKING, List, Dict

from tealer.teal.context.block_transaction_context import BlockTransactionContext

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.teal import Teal
    from tealer.teal.subroutine import Subroutine


# pylint: disable=too-few-public-methods,too-many-instance-attributes
class Function:
    """Class to represent a function in a contract."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        function_name: str,
        entry: "BasicBlock",
        blocks: List["BasicBlock"],
        contract: "Teal",
        main: "Subroutine",
        subroutines: Dict[str, "Subroutine"],
    ) -> None:
        self.function_name: str = function_name
        self.entry: "BasicBlock" = entry
        self._blocks: List["BasicBlock"] = blocks
        self.contract: "Teal" = contract
        self.main: "Subroutine" = main
        self.subroutines: Dict[str, "Subroutine"] = subroutines
        # A subroutine is represented using a single CFG. A subroutine can be part of multiple
        # functions. Basic Blocks part of the subroutine CFG will be part of all functions the subroutine
        # is used in. But, block transaction context is specific for each function. As a result,
        # We cannot store the block transaction context in basic block object.
        self._transaction_contexts: Dict["BasicBlock", "BlockTransactionContext"] = {
            block: BlockTransactionContext() for block in self._blocks
        }
        # caller blocks of a subroutine depend on the function __main__ CFG. As a result, they
        # cannot be computed without considering the function.
        self._subroutine_caller_blocks: Dict["Subroutine", List["BasicBlock"]] = {
            sub: [] for sub in subroutines.values()
        }

        for block in self._blocks:
            if block.is_callsub_block:
                self._subroutine_caller_blocks[block.called_subroutine].append(block)

        self._subroutine_return_point_blocks: Dict["Subroutine", List["BasicBlock"]] = {}
        for sub, caller_blocks in self._subroutine_caller_blocks.items():
            self._subroutine_return_point_blocks[sub] = [
                bi.next[0] for bi in caller_blocks if len(bi.next) == 1
            ]

        # print(self._subroutine_caller_blocks)

        # TODO: add exit_blocks

    @property
    def blocks(self) -> List["BasicBlock"]:
        return self._blocks

    def transaction_context(self, block: "BasicBlock") -> "BlockTransactionContext":
        return self._transaction_contexts[block]

    def caller_blocks(self, subroutine: "Subroutine") -> List["BasicBlock"]:
        """BasicBlock with callsub instructions which call the subroutine.

        Args:
            subroutine: the subroutine

        Returns:
            List of caller callsub basic blocks. This blocks have "callsub {subroutine.name}"
            as exit instruction.
        """
        return self._subroutine_caller_blocks[subroutine]

    def return_point_blocks(self, subroutine: "Subroutine") -> List["BasicBlock"]:
        return self._subroutine_return_point_blocks[subroutine]



"""Defines classes to represent asset_params_get fields.

``asset_params_get`` instruction is used to access parameter fields
of an asset.

Each field that can be accessed using asset_params_get is represented
by a class in tealer. All the classes representing the fields must inherit
from AssetParamsField class.

"""

# pylint: disable=too-few-public-methods
class AssetParamsField:
    """Base class to represent asset_params_get field."""

    def __init__(self) -> None:
        self._version: int = 2

    @property
    def version(self) -> int:
        """Teal version this field is introduced in and supported from.

        Returns:
            Teal version the field is introduced in and supported from.
        """
        return self._version

    def __str__(self) -> str:
        return self.__class__.__qualname__


class AssetTotal(AssetParamsField):
    """Total number of units of this asset."""


class AssetDecimals(AssetParamsField):
    """Number of digits to display after the decimal place."""


class AssetDefaultFrozen(AssetParamsField):
    """Whether the asset is frozen by default or not."""


class AssetUnitName(AssetParamsField):
    """Unit name of the asset."""


class AssetName(AssetParamsField):
    """The asset name."""


class AssetURL(AssetParamsField):
    """Url associated with the asset."""


class AssetMetadataHash(AssetParamsField):
    """32 byte commitment to some unspecified asset metadata."""


class AssetManager(AssetParamsField):
    """Manager address of the asset.

    Manager account is the only account that can authorize transactions
    to re-configure or destroy an asset.
    """


class AssetReserve(AssetParamsField):
    """Reserve address of the asset.

    Non-minted assets will reside in Reserve address instead of creator
    address if specified.
    """


class AssetFreeze(AssetParamsField):
    """Freeze address of the asset.

    The freeze account is allowed to freeze or unfreeze the asset holdings
    for a specific account.
    """


class AssetClawback(AssetParamsField):
    """Clawback address of the asset.

    The clawback address represents an account that is allowed to transfer
    assets from and to any asset holder.
    """


class AssetCreator(AssetParamsField):
    """Creator address of this asset."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 5



from typing import TYPE_CHECKING, List, Set, Tuple, Dict

from tealer.analyses.dataflow.transaction_context.generic import DataflowTransactionContext
from tealer.analyses.dataflow.transaction_context.utils.key_helpers import (
    get_gtxn_at_index_key,
    get_ind_base_for_gtxn_type_keys,
    is_value_matches_key,
    get_absolute_index_key,
    get_relative_index_key,
)
from tealer.teal.instructions.instructions import (
    Eq,
    Neq,
    Not,
)
from tealer.teal.instructions.transaction_field import (
    ApplicationID,
    TypeEnum,
    OnCompletion,
)
from tealer.utils.teal_enums import TealerTransactionType
from tealer.utils.teal_enums import (
    ALL_TRANSACTION_TYPES,
    APPLICATION_TRANSACTION_TYPES,
    TYPEENUM_TRANSACTION_TYPES,
)
from tealer.utils.teal_enums import oncompletion_to_tealer_type, transaction_type_to_tealer_type
from tealer.utils.analyses import is_int_push_ins
from tealer.analyses.utils.stack_ast_builder import KnownStackValue, UnknownStackValue
from tealer.utils.algorand_constants import MAX_GROUP_SIZE

if TYPE_CHECKING:
    from tealer.teal.instructions.instructions import Instruction
    from tealer.teal.instructions.transaction_field import TransactionField
    from tealer.teal.basic_blocks import BasicBlock

# TODO: change the value representation for TxnTyp to something similar to FeeValue.
# Same underlying type(TealerTransactionType) for three different things:
#   - CallConfig: (Creation(AppId == 0) or Call(AppId != 0) or Always)
#   - TypeEnum: (pay or keyreg or axfer or afrz or appl)
#   - OnCompletion: (NoOp, OptIn, CloseOut, ClearState, UpdateApplication or DeleteApplication)

transaction_type_key = "TransactionType"

base_keys = [transaction_type_key]
universal_sets: Dict[str, List] = {}
universal_sets[transaction_type_key] = list(ALL_TRANSACTION_TYPES)


class TxnType(DataflowTransactionContext):  # pylint: disable=too-few-public-methods

    TRANSACTION_TYPE_KEY = transaction_type_key
    BASE_KEYS = base_keys
    KEYS_WITH_GTXN: List[str] = [transaction_type_key]
    UNIVERSAL_SETS: Dict[str, List] = universal_sets

    def _universal_set(self, key: str) -> Set:
        if key not in self.BASE_KEYS:
            _, base_key = get_ind_base_for_gtxn_type_keys(key)
            return set(self.UNIVERSAL_SETS[base_key])
        return set(self.UNIVERSAL_SETS[key])

    def _null_set(self, key: str) -> Set:
        return set()

    def _union(self, key: str, a: Set, b: Set) -> Set:
        return a | b

    def _intersection(self, key: str, a: Set, b: Set) -> Set:
        return a & b

    def _get_asserted_transaction_types(  # pylint: disable=too-many-branches, too-many-locals
        self, key: str, ins_stack_value: KnownStackValue
    ) -> Tuple[Set["TealerTransactionType"], Set["TealerTransactionType"]]:
        """return set of transaction type that make the comparison true and a set that makes it false.

        considered instruction patterns:
        [g]txn [idx] ApplicationID
        bz/bnz

        [g]txn [idx] ApplicationID
        !
        bz/bnz

        [g]txn [idx] ApplicationID
        int 0
        ( == | != )
        bz/bnz

        [g]txn [idx] TypeEnum
        int [pay | acfg | axfer | appl ..]
        ( == | != )
        bz/bnz

        [g]txn [idx] OnCompletion
        int [UpdateApplication | DeleteApplication | ...]
        ( == | != )
        bz/bnz

        Args:
            key: The analysis key.
            ins_stack_value: A stack value. The stack value represents the result of ins_stack_value.instruction.
                This function assumes that this value is being asserted/checked.

        Returns:
            Set of possible values for the :key: when the ins_stack_value is asserted to be True and when ins_stack_value is
            asserted to be False.
        """
        U = set(self.UNIVERSAL_SETS[self.TRANSACTION_TYPE_KEY])

        ins1 = ins_stack_value.instruction
        if is_value_matches_key(key, ins_stack_value, ApplicationID):
            # txn ApplicationID pushes 0 if this Application creation transaction elses pushes nonzero
            return set(APPLICATION_TRANSACTION_TYPES) - set(
                [TealerTransactionType.ApplCreation]
            ), set([TealerTransactionType.ApplCreation])

        if isinstance(ins1, Not):
            not_arg = ins_stack_value.args[0]
            if isinstance(not_arg, UnknownStackValue):
                return set(U), set(U)
            ins2 = not_arg.instruction
            if is_value_matches_key(key, not_arg, ApplicationID):
                # txn ApplicationID
                # !
                # return
                return set([TealerTransactionType.ApplCreation]), set(
                    APPLICATION_TRANSACTION_TYPES
                ) - set([TealerTransactionType.ApplCreation])
            return set(U), set(U)

        if isinstance(ins1, (Eq, Neq)):
            arg1 = ins_stack_value.args[0]
            arg2 = ins_stack_value.args[1]
            if isinstance(arg1, UnknownStackValue) or isinstance(arg2, UnknownStackValue):
                # if one of the args is unknown
                return set(U), set(U)
            ins2 = arg1.instruction
            ins3 = arg2.instruction

            is_int_ins2, value_2 = is_int_push_ins(ins2)
            is_int_ins3, value_3 = is_int_push_ins(ins3)

            if (is_int_ins2 and is_int_ins3) or ((not is_int_ins2) and (not is_int_ins3)):
                # if both are ints or both are not ints
                return set(U), set(U)

            true_values, false_values = None, None
            if is_value_matches_key(key, arg1, ApplicationID) and value_3 is not None:
                # TODO: ApplicationCreation transaction can be NoOp or OptIn
                if isinstance(value_3, int) and value_3 == 0:
                    true_values, false_values = set([TealerTransactionType.ApplCreation]), set(
                        APPLICATION_TRANSACTION_TYPES
                    ) - set([TealerTransactionType.ApplCreation])
            elif is_value_matches_key(key, arg2, ApplicationID) and value_2 is not None:
                # TODO: ApplicationCreation transaction can be NoOp or OptIn
                if isinstance(value_2, int) and value_2 == 0:
                    true_values, false_values = set([TealerTransactionType.ApplCreation]), set(
                        APPLICATION_TRANSACTION_TYPES
                    ) - set([TealerTransactionType.ApplCreation])

            if is_value_matches_key(key, arg1, TypeEnum) and value_3 is not None:
                compared_type = transaction_type_to_tealer_type(value_3)
                true_values, false_values = set([compared_type]), set(
                    TYPEENUM_TRANSACTION_TYPES
                ) - set([compared_type])
            elif is_value_matches_key(key, arg2, TypeEnum) and value_2 is not None:
                compared_type = transaction_type_to_tealer_type(value_2)
                true_values, false_values = set([compared_type]), set(
                    TYPEENUM_TRANSACTION_TYPES
                ) - set([compared_type])

            if is_value_matches_key(key, arg1, OnCompletion) and value_3 is not None:
                compared_on_completion = oncompletion_to_tealer_type(value_3)
                true_values, false_values = set([compared_on_completion]), set(
                    APPLICATION_TRANSACTION_TYPES
                ) - set([compared_on_completion])
            elif is_value_matches_key(key, arg2, OnCompletion) and value_2 is not None:
                compared_on_completion = oncompletion_to_tealer_type(value_2)
                true_values, false_values = set([compared_on_completion]), set(
                    APPLICATION_TRANSACTION_TYPES
                ) - set([compared_on_completion])

            if true_values is not None and false_values is not None:
                if isinstance(ins1, Eq):
                    return true_values, false_values
                return false_values, true_values

        return set(U), set(U)

    def _get_asserted_single(self, key: str, ins_stack_value: KnownStackValue) -> Tuple[Set, Set]:
        return self._get_asserted_transaction_types(key, ins_stack_value)

    def _store_results(self) -> None:
        transaction_type_context = self._block_contexts[self.TRANSACTION_TYPE_KEY]

        for block in self._function.blocks:
            self._function.transaction_context(block).transaction_types = list(
                transaction_type_context[block]
            )

            for idx in range(MAX_GROUP_SIZE):
                values = self._block_contexts[
                    get_gtxn_at_index_key(idx, self.TRANSACTION_TYPE_KEY)
                ][block]
                self._function.transaction_context(block).gtxn_context(
                    idx
                ).transaction_types = list(values)

                abs_values = self._block_contexts[
                    get_absolute_index_key(idx, self.TRANSACTION_TYPE_KEY)
                ][block]
                self._function.transaction_context(block).absolute_context(
                    idx
                ).transaction_types = list(abs_values)

            for offset in range(-(MAX_GROUP_SIZE - 1), MAX_GROUP_SIZE):
                if offset == 0:
                    continue
                rel_values = self._block_contexts[
                    get_relative_index_key(offset, self.TRANSACTION_TYPE_KEY)
                ][block]
                self._function.transaction_context(block).relative_context(
                    offset
                ).transaction_types = list(rel_values)



"""Contains functions, classes for parsing and representing teal contracts.

Module contains implementations for representing and parsing teal contracts.

Additional to instructions, this modules contains implementation of Teal, BasicBlock
classes. BasicBlock class is used to represent a basic block in the teal contract.
Teal class is used to represent the teal contract. Teal class also comes with
api methods to run detectors, printers on the contract.

Modules:
    instructions: Contains functions, classes for parsing and representing teal
        instructions.

    basic_blocks: Defines class to represent basic blocks of teal contracts.

    global_field: Defines classes to represent global fields.

    parse_teal: Defines parsing functions for teal contracts.

    teal: Defines class to represent Teal contracts.
"""



"""Parser for fields of acct_params_get.

Each ``acct_params_get`` field is represented as a class. Parsing
the field is creating the class instance representing the field given
it's string representation.

Fields of ``acct_params_get`` doesn't have immediate arguments
and have very simple structure. All fields map one-to-one with
their string representation. As a result, app parameter fields are
parsed by first constructing a map(dict) from their string
representation to corresponding class representing the field and
using that to find the correct class for the given field.

Attributes:
    ACCT_PARAMS_FIELD_TXT_TO_OBJECT: Map(dict) from string representation
        of app parameter field to the corresponding class.
"""

from tealer.teal.instructions.acct_params_field import (
    AcctParamsField,
    AcctBalance,
    AcctMinBalance,
    AcctAuthAddr,
    AcctTotalNumUint,
    AcctTotalNumByteSlice,
    AcctTotalExtraAppPages,
    AcctTotalAppsCreated,
    AcctTotalAppsOptedIn,
    AcctTotalAssetsCreated,
    AcctTotalAssets,
    AcctTotalBoxes,
    AcctTotalBoxBytes,
)


ACCT_PARAMS_FIELD_TXT_TO_OBJECT = {
    "AcctBalance": AcctBalance,
    "AcctMinBalance": AcctMinBalance,
    "AcctAuthAddr": AcctAuthAddr,
    "AcctTotalNumUint": AcctTotalNumUint,
    "AcctTotalNumByteSlice": AcctTotalNumByteSlice,
    "AcctTotalExtraAppPages": AcctTotalExtraAppPages,
    "AcctTotalAppsCreated": AcctTotalAppsCreated,
    "AcctTotalAppsOptedIn": AcctTotalAppsOptedIn,
    "AcctTotalAssetsCreated": AcctTotalAssetsCreated,
    "AcctTotalAssets": AcctTotalAssets,
    "AcctTotalBoxes": AcctTotalBoxes,
    "AcctTotalBoxBytes": AcctTotalBoxBytes,
}


def parse_acct_params_field(field: str) -> AcctParamsField:
    """Parse fields of ``app_params_get``.

    Args:
        field: string representation of the field.

    Returns:
        object of class corresponding to the given app_params_get field.
    """

    return ACCT_PARAMS_FIELD_TXT_TO_OBJECT[field]()



"""Util functions to compute code complexity of the contracts.

Module implements functions to calculate cyclomatic complexity of
the contract for now.

Functions:
    compute_cyclomatic_complexity(cfg: List["BasicBlock"]) -> int:
        Calculates the cyclomatic complexity of the contract given
        it's Control Flow Graph(CFG) :cfg:.

"""
# pylint: skip-file
# mypy: ignore-errors
# TODO: Functions in this module are not used anywhere. Have to decide on how to calculate the code complexity.

from typing import List, TYPE_CHECKING
from tealer.utils.analyses import next_blocks_global, prev_blocks_global

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock


# from slither: slither/utils/code_complexity.py
def _compute_number_edges(cfg: List["BasicBlock"]) -> int:
    """compute number of edges in the control flow graph of the contract.

    Args:
        cfg: Control Flow Graph(CFG) of the contract.

    Returns:
        number of edges in the CFG.
    """

    n = 0
    for bb in cfg:
        n += len(next_blocks_global(bb))
    return n


# from slither: slither/utils/code_complexity.py
def _compute_strongly_connected_components(cfg: List["BasicBlock"]) -> List[List["BasicBlock"]]:
    """Compute strongly connected components in the control flow graph.

    Implementation uses Kosaraju algorithm to find strongly connected
    components in the control flow graph of the contract. This follows
    the implementation described in the `Kosaraju algorithm wikipedia page`_.

    Args:
        cfg : Control Flow Graph(CFG) of the contract

    Returns:
        list of strongly connected components in CFG. each connected component
        is a list of basic blocks(nodes in CFG).

    .. _Kosaraju algorithm wikipedia page:
       https://en.wikipedia.org/wiki/Kosaraju%27s_algorithm#The_algorithm
    """

    visited = {bb: False for bb in cfg}
    assigned = {bb: False for bb in cfg}
    components = []

    l = []

    def visit(bb: "BasicBlock") -> None:
        if not visited[bb]:
            visited[bb] = True
            for next_bb in next_blocks_global(bb):
                visit(next_bb)
            l.append(bb)

    for bb in cfg:
        visit(bb)

    def assign(bb: "BasicBlock", root: List["BasicBlock"]) -> None:
        if not assigned[bb]:
            assigned[bb] = True
            root.append(bb)
            for prev_bb in prev_blocks_global(bb):
                assign(prev_bb, root)

    for bb in l:
        component: List["BasicBlock"] = []
        assign(bb, component)
        if component:
            components.append(component)

    return components


# from slither: slither/utils/code_complexity.py
def compute_cyclomatic_complexity(cfg: List["BasicBlock"]) -> int:
    """Compute cyclomatic complexity of the contract.

    Args:
        cfg: Control Flow Graph(CFG) of the contract.

    Returns:
        Cyclomatic complexity of the contract.
    """

    # from https://en.wikipedia.org/wiki/Cyclomatic_complexity
    # M = E - N + 2P
    # where M is the complexity
    # E number of edges
    # N number of nodes
    # P number of connected components

    E = _compute_number_edges(cfg)
    N = len(cfg)
    P = len(_compute_strongly_connected_components(cfg))
    return E - N + 2 * P



import re
from pathlib import Path
from typing import List, Optional, Tuple, Set

from tealer.teal.instructions.instructions import Instruction, Label
from tealer.teal.instructions.parse_instruction import parse_line
from tealer.teal.teal import Teal
from tealer.utils.output import CFGDotConfig, full_cfg_to_dot


# pylint: disable=too-few-public-methods
class Regex:
    def __init__(self, label: str, instructions: List[Instruction]) -> None:
        self.label = label
        self.instructions = instructions


def parse_regex(regex: str) -> Regex:
    """
    Parse the regex

    Args:
        regex: regex as a string

    Returns:
        Regex as an object

    Raises:
        ValueError: if the text does not have the correct format
    """
    pattern = re.compile(r"^(.*?)=>\s*(.*?)$", re.DOTALL)
    match = pattern.search(regex)
    if match:
        # Group 1 is the label, Group 2 is the content
        label_found, content_found = match.groups()
        label = label_found.strip()
        instructions_text = content_found.strip().splitlines()

        instructions: List[Instruction] = []
        for line in instructions_text:
            instruction = parse_line(line)
            if instruction:
                instructions.append(instruction)

        return Regex(label, instructions)

    raise ValueError("Invalid text format")


def _find_label(instructions: List[Instruction], label: str) -> Optional[Instruction]:
    # Assuming "*" is not a valid teal label ?
    # Assuming the first instruction is the entry point
    if label == "*":
        return instructions[0]

    match = [ins for ins in instructions if isinstance(ins, Label) and ins.label == label]

    # assuming the same label can't be reused in a contract
    assert len(match) <= 1

    if match:
        return match[0]

    return None


def _is_equal(a: Instruction, b: Instruction) -> bool:
    if type(a) is not type(b):
        return False

    # We don't have current a good way to compare two instructions subfield
    # So we use the string representation as a temp solution
    # Note both objects are not expected to be the same, as one is coming from the teal contract
    # The other one was generated based on the regex
    return str(a) == str(b)


def _is_match(current_instruction: Optional[Instruction], regex: List[Instruction]) -> bool:
    for _, regex_ins in enumerate(regex):
        if current_instruction is None:
            return False
        if not _is_equal(current_instruction, regex_ins):
            return False
        current_instruction = (
            current_instruction.next[0]
            if current_instruction.next and len(current_instruction.next) == 1
            else None
        )

    return True


def _find_instructions(
    current_instruction: Instruction,
    regex: List[Instruction],
    visited: Set[Instruction],
    matches: List[List[Instruction]],
    covered: Set[Instruction],
) -> bool:
    if current_instruction in visited:
        return False

    visited.add(current_instruction)

    reaches = False

    if _is_match(current_instruction, regex):
        match: List[Instruction] = []
        current = current_instruction
        for _ in range(0, len(regex) - 1):

            if not current:
                break
            match.append(current)

            if len(current.next) != 1:
                print(f"Regex cannot work on branching instructions {current_instruction}")
                return False
            current = current.next[0]

        match.append(current)
        matches.append(match)
        reaches = True

    for next_ins in current_instruction.next:

        if next_ins in covered:
            continue

        if _find_instructions(next_ins, regex, visited, matches, covered):
            covered.add(current_instruction)
            reaches = True

    return reaches


def match_regex(contract: Teal, regex: Regex) -> Tuple[List[List[Instruction]], Set[Instruction]]:
    """
    Match the regex in the contract

    Args:
        contract: target
        regex: regex

    Returns:
        A tuple of (matches, covered), where
        - matches: contains all the instructions that match the regex
        - covered: contains all the instruction between the label and the matches
    """
    label = _find_label(contract.instructions, regex.label)

    if not label:
        print(f"{label} cannot be found in {contract}")
        return [], set()

    matches: List[List[Instruction]] = []
    covered: Set[Instruction] = set()

    _find_instructions(label, regex.instructions, set(), matches, covered)

    return matches, covered


def update_config(
    config: CFGDotConfig, matches: List[List[Instruction]], covered: Set[Instruction]
) -> None:
    """
    Update the config with the regex result. It color match in red, and covered in green

    Args:
        config: config
        matches: instructions that match the regex
        covered: instructions between the label and the match
    """
    for inss in matches:
        for ins in inss:
            config.custom_background_color[ins] = "#e0182b"  # red

    for ins in covered:
        config.custom_background_color[ins] = "#36d899"  # green


def run_regex(teal: Teal, regex_Path: Path, export_path: Path) -> None:
    """
    Run the regex analysis and generate the dot file

    Args:
        teal: target
        regex_Path: path to the regex
        export_path: path to generate the dot file
    """
    with open(regex_Path, "r", encoding="utf8") as f_regex:
        regex = parse_regex(f_regex.read())

    matches, covered = match_regex(teal, regex)

    if not matches:
        print("Not match was found")
        return
    # Ensure the CFG generation does not crash
    config = CFGDotConfig()
    update_config(config, matches, covered)

    with open(export_path, "w", encoding="utf8") as fp:
        full_cfg_to_dot(teal, config, Path(fp.name))

    print(f"Result generated in {export_path}")



from typing import List, Tuple
import pytest

from tealer.utils.teal_enums import TealerTransactionType
from tealer.utils.algorand_constants import MAX_GROUP_SIZE
from tealer.utils.command_line.common import init_tealer_from_single_contract

from tests.utils import order_basic_blocks


CAN_UPDATE = """
#pragma version 2
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    txn OnCompletion
    int OptIn
    ==
    bnz handle_optin
    txn OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    txn OnCompletion
    int DeleteApplication
    ==
    bnz handle_deleteapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_deleteapp:
err
"""

CAN_UPDATE_TX_TYPES = [
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplCreation,
    ],
    [TealerTransactionType.ApplCreation],
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
    ],
    [
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
    ],
    [
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
    ],
    [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
    [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
    [TealerTransactionType.ApplNoOp],
    [TealerTransactionType.ApplNoOp, TealerTransactionType.ApplOptIn],
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
    ],
    [],
]

CAN_UPDATE_LOOP = """
#pragma version 4
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    txn OnCompletion
    int OptIn
    ==
    bnz handle_optin
    txn OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    b loop
handle_noop:
handle_optin:
handle_closeout:
int 1
return
loop:
    txn OnCompletion
    int DeleteApplication
    !=
    assert
    load 0
    int 5
    <
    bz end
    load 0
    int 1
    +
    store 0
    b loop
end:
    int 1
    return
"""

CAN_UPDATE_LOOP_TX_TYPES = [
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplCreation,
    ],
    [TealerTransactionType.ApplCreation],
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
    ],
    [
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
    ],
    [
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
    ],
    [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
    [TealerTransactionType.ApplNoOp],
    [TealerTransactionType.ApplNoOp, TealerTransactionType.ApplOptIn],
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
    ],
    [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
    [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
    [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
]

SUBROUTINE = """
#pragma version 5
b main
push_zero:
    int 0
    retsub
is_even:
    int 2
    %
    bz return_1
    callsub push_zero
    retsub
return_1:
    int 1
    retsub
main:
    int 4
    bz path_1
    b path_2
path_1:
   txn OnCompletion
   int UpdateApplication
   ==
   assert
   callsub is_even
   int 1
   return
path_2:
   txn OnCompletion
   int DeleteApplication
   ==
   assert
   callsub is_even
   int 1
   int 1
   return
"""

SUBROUTINE_TX_TYPES = [
    [TealerTransactionType.ApplUpdateApplication, TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplUpdateApplication, TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplUpdateApplication, TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplUpdateApplication, TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplUpdateApplication, TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplUpdateApplication, TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplUpdateApplication, TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplUpdateApplication],
    [TealerTransactionType.ApplUpdateApplication],
    [TealerTransactionType.ApplDeleteApplication],
    [TealerTransactionType.ApplDeleteApplication],
]


CAN_UPDATE_GTXN_0 = """
#pragma version 2
int 0
gtxn 0 ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    gtxn 0 OnCompletion
    int DeleteApplication
    ==
    bnz handle_deleteapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_deleteapp:
err
"""

CAN_UPDATE_LOOP_GTXN_0 = """
#pragma version 4
int 0
gtxn 0 ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    b loop
handle_noop:
handle_optin:
handle_closeout:
int 1
return
loop:
    gtxn 0 OnCompletion
    int DeleteApplication
    !=
    assert
    load 0
    int 5
    <
    bz end
    load 0
    int 1
    +
    store 0
    b loop
end:
    int 1
    return
"""


ALL_TESTS_TXN = [
    (CAN_UPDATE, CAN_UPDATE_TX_TYPES, -1),
    (CAN_UPDATE_LOOP, CAN_UPDATE_LOOP_TX_TYPES, -1),
    (SUBROUTINE, SUBROUTINE_TX_TYPES, -1),
    (CAN_UPDATE_GTXN_0, CAN_UPDATE_TX_TYPES, 0),
    (CAN_UPDATE_LOOP_GTXN_0, CAN_UPDATE_LOOP_TX_TYPES, 0),
]


@pytest.mark.parametrize("test", ALL_TESTS_TXN)  # type: ignore
def test_tx_types(test: Tuple[str, List[List[int]], int]) -> None:
    code, tx_types_list, idx = test

    tealer = init_tealer_from_single_contract(code.strip(), "test")
    function = tealer.contracts["test"].functions["test"]

    bbs = order_basic_blocks(function.blocks)
    for b, tx_types in zip(bbs, tx_types_list):
        if idx == -1:
            ctx = function.transaction_context(b)
        else:
            ctx = function.transaction_context(b).gtxn_context(idx)
        assert set(ctx.transaction_types) == set(tx_types)


TEST_GROUP_INDICES = """
#pragma version 6
global GroupSize
int 1
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    gtxn 0 OnCompletion
    int DeleteApplication
    ==
    bnz handle_deleteapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_deleteapp:
err
"""

txn_types_list = [
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
        TealerTransactionType.ApplCreation,
    ],
    [TealerTransactionType.ApplCreation],
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
    ],
    [
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
    ],
    [
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
    ],
    [
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
    ],
    [
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
    ],
    [TealerTransactionType.ApplNoOp],
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
    ],
    [
        TealerTransactionType.ApplNoOp,
        TealerTransactionType.ApplOptIn,
        TealerTransactionType.ApplCloseOut,
        TealerTransactionType.ApplClearState,
        TealerTransactionType.ApplUpdateApplication,
        TealerTransactionType.ApplDeleteApplication,
    ],
    [],
]

# { for each index: {
#           for each block: {
#               possible transaction types
#       }
# }
gtxn_types_list: List[List[List[TealerTransactionType]]] = [
    [
        [
            TealerTransactionType.ApplNoOp,
            TealerTransactionType.ApplOptIn,
            TealerTransactionType.ApplCloseOut,
            TealerTransactionType.ApplClearState,
            TealerTransactionType.ApplUpdateApplication,
            TealerTransactionType.ApplCreation,
        ],
        [TealerTransactionType.ApplCreation],
        [
            TealerTransactionType.ApplNoOp,
            TealerTransactionType.ApplOptIn,
            TealerTransactionType.ApplCloseOut,
            TealerTransactionType.ApplClearState,
            TealerTransactionType.ApplUpdateApplication,
        ],
        [
            TealerTransactionType.ApplOptIn,
            TealerTransactionType.ApplCloseOut,
            TealerTransactionType.ApplClearState,
            TealerTransactionType.ApplUpdateApplication,
        ],
        [
            TealerTransactionType.ApplCloseOut,
            TealerTransactionType.ApplClearState,
            TealerTransactionType.ApplUpdateApplication,
        ],
        [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
        [TealerTransactionType.ApplClearState, TealerTransactionType.ApplUpdateApplication],
        [TealerTransactionType.ApplNoOp],
        [TealerTransactionType.ApplNoOp, TealerTransactionType.ApplOptIn],
        [
            TealerTransactionType.ApplNoOp,
            TealerTransactionType.ApplOptIn,
            TealerTransactionType.ApplCloseOut,
        ],
        [],
    ],
]

gtxn_types_list += [[[]] * 11 for _ in range(MAX_GROUP_SIZE - 1)]


ALL_TESTS_GTXN = [
    (TEST_GROUP_INDICES, txn_types_list, gtxn_types_list),
]


@pytest.mark.parametrize("test", ALL_TESTS_GTXN)  # type: ignore
def test_tx_types_gtxn(
    test: Tuple[
        str,
        List[List[TealerTransactionType]],
        List[List[List[TealerTransactionType]]],
    ],
) -> None:
    code, ex_txn_types_list, ex_gtxn_types_list = test

    tealer = init_tealer_from_single_contract(code.strip(), "test")
    function = tealer.contracts["test"].functions["test"]

    bbs = order_basic_blocks(function.blocks)
    print("number of blocks:", len(bbs))
    for block_num, b in enumerate(bbs):
        print(block_num, b)
        assert set(function.transaction_context(b).transaction_types) == set(
            ex_txn_types_list[block_num]
        )
        print(block_num, b)
        for txn_ind in range(MAX_GROUP_SIZE):
            ctx = function.transaction_context(b).gtxn_context(txn_ind)
            print("txn_ind =", txn_ind)
            assert set(ctx.transaction_types) == set(ex_gtxn_types_list[txn_ind][block_num])



from typing import List, Tuple, Dict
import pytest


from tealer.teal.basic_blocks import BasicBlock
from tealer.teal.subroutine import Subroutine
from tealer.teal.instructions import instructions
from tealer.teal.instructions import transaction_field
from tealer.teal.parse_teal import parse_teal

from tests.utils import cmp_cfg, construct_cfg


MULTIPLE_RETSUB = """
#pragma version 5
b main
is_even:
    int 2
    %
    bz return_1
    int 0
    retsub
return_1:
    int 1
    retsub
main:
    int 4
    callsub is_even
    return
"""

ins_list = [
    instructions.Pragma(5),
    instructions.B("main"),
    instructions.Label("is_even"),
    instructions.Int(2),
    instructions.Modulo(),
    instructions.BZ("return_1"),
    instructions.Int(0),
    instructions.Retsub(),
    instructions.Label("return_1"),
    instructions.Int(1),
    instructions.Retsub(),
    instructions.Label("main"),
    instructions.Int(4),
    instructions.Callsub("is_even"),
    instructions.Return(),
]

ins_partitions = [(0, 2), (2, 6), (6, 8), (8, 11), (11, 14), (14, 15)]
bbs_links = [(0, 4), (4, 5), (1, 2), (1, 3)]

bbs_edges_new = [(0, 4), (4, 5), (1, 2), (1, 3)]


MULTIPLE_RETSUB_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)

MULTIPLE_RETSUB_CFG_NEW = construct_cfg(ins_list, ins_partitions, bbs_edges_new)

bbs = MULTIPLE_RETSUB_CFG_NEW
MULTIPLE_RETSUB_MAIN_NEW = Subroutine("__main__", bbs[0], [bbs[0], bbs[4], bbs[5]])
MULTIPLE_RETSUB_SUBROUTINES_NEW = {
    "is_even": Subroutine("is_even", bbs[1], [bbs[1], bbs[2], bbs[3]])
}

MULTIPLE_RETSUB_SUBROUTINES_NEW["is_even"].caller_blocks = [bbs[4]]

SUBROUTINE_BACK_JUMP = """
#pragma version 5
b main
getmod:
    %
    retsub
is_odd:
    int 2
    b getmod
main:
    int 5
    callsub is_odd
    return
"""

ins_list = [
    instructions.Pragma(5),
    instructions.B("main"),
    instructions.Label("getmod"),
    instructions.Modulo(),
    instructions.Retsub(),
    instructions.Label("is_odd"),
    instructions.Int(2),
    instructions.B("getmod"),
    instructions.Label("main"),
    instructions.Int(5),
    instructions.Callsub("is_odd"),
    instructions.Return(),
]

ins_partitions = [(0, 2), (2, 5), (5, 8), (8, 11), (11, 12)]
bbs_links = [(0, 3), (3, 4), (2, 1)]

bbs_edges_new = [(0, 3), (3, 4), (2, 1)]

SUBROUTINE_BACK_JUMP_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)
SUBROUTINE_BACK_JUMP_CFG_NEW = construct_cfg(ins_list, ins_partitions, bbs_edges_new)

bbs = SUBROUTINE_BACK_JUMP_CFG_NEW
SUBROUTINE_BACK_JUMP_MAIN_NEW = Subroutine("__main__", bbs[0], [bbs[0], bbs[3], bbs[4]])
SUBROUTINE_BACK_JUMP_SUBROUTINES_NEW = {"is_odd": Subroutine("is_odd", bbs[2], [bbs[1], bbs[2]])}
SUBROUTINE_BACK_JUMP_SUBROUTINES_NEW["is_odd"].caller_blocks = [bbs[3]]


BRANCHING = """
#pragma version 2
txn NumAppArgs
int 2
==
bz fin
byte "myarg"
txn ApplicationArgs 0
==
bnz check_second_arg
int 0
return
check_second_arg:
txn ApplicationArgs 1
btoi
int 100
>
bnz fin
int 0
return
fin:
int 1
return
"""

ins_list = [
    instructions.Pragma(2),
    instructions.Txn(transaction_field.NumAppArgs()),
    instructions.Int(2),
    instructions.Eq(),
    instructions.BZ("fin"),
    instructions.Byte('"myarg"'),
    instructions.Txn(transaction_field.ApplicationArgs(0)),
    instructions.Eq(),
    instructions.BNZ("check_second_arg"),
    instructions.Int(0),
    instructions.Return(),
    instructions.Label("check_second_arg"),
    instructions.Txn(transaction_field.ApplicationArgs(1)),
    instructions.Btoi(),
    instructions.Int(100),
    instructions.Greater(),
    instructions.BNZ("fin"),
    instructions.Int(0),
    instructions.Return(),
    instructions.Label("fin"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [(0, 5), (5, 9), (9, 11), (11, 17), (17, 19), (19, 22)]
bbs_links = [(0, 1), (0, 5), (1, 2), (1, 3), (3, 4), (3, 5)]

BRANCHING_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)
BRANCHING_CFG_NEW = construct_cfg(ins_list, ins_partitions, bbs_links)

bbs = BRANCHING_CFG_NEW
BRANCHING_MAIN_NEW = Subroutine(
    "__main__", bbs[0], [bbs[0], bbs[1], bbs[2], bbs[3], bbs[4], bbs[5]]
)
BRANCHING_SUBROUTINES_NEW: Dict[str, Subroutine] = {}


LOOPS = """
int 0
loop:
    dup
    int 5
    <
    bz end
    int 1
    +
    b loop
end:
    int 1
    return
"""

ins_list = [
    instructions.Int(0),
    instructions.Label("loop"),
    instructions.Dup(),
    instructions.Int(5),
    instructions.Less(),
    instructions.BZ("end"),
    instructions.Int(1),
    instructions.Add(),
    instructions.B("loop"),
    instructions.Label("end"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [(0, 1), (1, 6), (6, 9), (9, 12)]
bbs_links = [(0, 1), (1, 2), (1, 3), (2, 1)]

LOOPS_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)
LOOPS_CFG_NEW = construct_cfg(ins_list, ins_partitions, bbs_links)

bbs = LOOPS_CFG_NEW
LOOPS_MAIN_NEW = Subroutine("__main__", bbs[0], [bbs[0], bbs[1], bbs[2], bbs[3]])
LOOPS_SUBROUTINES_NEW: Dict[str, Subroutine] = {}


SWITCH_N_MATCH = """
#pragma version 8
int 1
switch label1 label2
int 1
int 2
int 1
match label1 label2
int 0
return
label1:
int 2
return
label2:
int 3
return
"""

ins_list = [
    instructions.Pragma(8),
    instructions.Int(1),
    instructions.Switch(["label1", "label2"]),
    instructions.Int(1),
    instructions.Int(2),
    instructions.Int(1),
    instructions.Match(["label1", "label2"]),
    instructions.Int(0),
    instructions.Return(),
    instructions.Label("label1"),
    instructions.Int(2),
    instructions.Return(),
    instructions.Label("label2"),
    instructions.Int(3),
    instructions.Return(),
]

ins_partitions = [(0, 3), (3, 7), (7, 9), (9, 12), (12, 15)]
bbs_links = [(0, 1), (0, 3), (0, 4), (1, 2), (1, 3), (1, 4)]

SWITCH_N_MATCH_CFG = construct_cfg(ins_list, ins_partitions, bbs_links)
SWITCH_N_MATCH_CFG_NEW = construct_cfg(ins_list, ins_partitions, bbs_links)

bbs = SWITCH_N_MATCH_CFG_NEW
SWITCH_N_MATCH_MAIN_NEW = Subroutine("__main__", bbs[0], [bbs[0], bbs[1], bbs[2], bbs[3], bbs[4]])
SWITCH_N_MATCH_SUBROUTINES_NEW: Dict[str, Subroutine] = {}


ALL_TESTS = [
    (MULTIPLE_RETSUB, MULTIPLE_RETSUB_CFG),
    (SUBROUTINE_BACK_JUMP, SUBROUTINE_BACK_JUMP_CFG),
    (BRANCHING, BRANCHING_CFG),
    (LOOPS, LOOPS_CFG),
    (SWITCH_N_MATCH, SWITCH_N_MATCH_CFG),
]

ALL_TESTS_NEW = [
    (
        MULTIPLE_RETSUB,
        MULTIPLE_RETSUB_CFG_NEW,
        MULTIPLE_RETSUB_MAIN_NEW,
        MULTIPLE_RETSUB_SUBROUTINES_NEW,
    ),
    (
        SUBROUTINE_BACK_JUMP,
        SUBROUTINE_BACK_JUMP_CFG_NEW,
        SUBROUTINE_BACK_JUMP_MAIN_NEW,
        SUBROUTINE_BACK_JUMP_SUBROUTINES_NEW,
    ),
    (BRANCHING, BRANCHING_CFG_NEW, BRANCHING_MAIN_NEW, BRANCHING_SUBROUTINES_NEW),
    (LOOPS, LOOPS_CFG_NEW, LOOPS_MAIN_NEW, LOOPS_SUBROUTINES_NEW),
    (
        SWITCH_N_MATCH,
        SWITCH_N_MATCH_CFG_NEW,
        SWITCH_N_MATCH_MAIN_NEW,
        SWITCH_N_MATCH_SUBROUTINES_NEW,
    ),
]


@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_cfg_construction(test: Tuple[str, List[BasicBlock]]) -> None:
    code, cfg = test
    teal = parse_teal(code.strip())
    for bb in cfg:
        print(bb)
    for bb in teal.bbs:
        print(bb)
    print("*" * 20)
    assert cmp_cfg(teal.bbs, cfg)


@pytest.mark.parametrize("test", ALL_TESTS_NEW)  # type: ignore
def test_cfg_construction_new(
    test: Tuple[str, List[BasicBlock], Subroutine, Dict[str, Subroutine]]
) -> None:
    code, expected_cfg, expected_main, expected_subroutines = test
    teal = parse_teal(code.strip())
    for bb in expected_cfg:
        print(bb)
    for bb in teal.bbs:
        print(bb)
    print("*" * 20)
    assert cmp_cfg(teal.bbs, expected_cfg)
    test_main = teal.main
    test_subroutines = teal.subroutines

    assert cmp_cfg([test_main.entry], [expected_main.entry])
    assert cmp_cfg(test_main.blocks, expected_main.blocks)

    assert len(test_subroutines.keys()) == len(expected_subroutines.keys())
    assert sorted(test_subroutines.keys()) == sorted(expected_subroutines.keys())
    for ex_name in expected_subroutines:
        test_subroutine = test_subroutines[ex_name]
        expected_subroutine = expected_subroutines[ex_name]
        assert test_subroutine.name == expected_subroutine.name
        assert cmp_cfg([test_subroutine.entry], [expected_subroutine.entry])
        assert cmp_cfg(test_subroutine.blocks, expected_subroutine.blocks)
        assert cmp_cfg(test_subroutine.caller_blocks, expected_subroutine.caller_blocks)
    assert len(test_subroutines) == len(expected_subroutines)
